#!/bin/sh
""""exec ${PYTHON:-python} -t $0 "$@";" """


__author__ = "Chad MILLER <lddaap-project % chad.org>"
__date__ = "2007-11-11"
__version__ = "0.1"

"""\
lddaap
======

Long distance DAAP

Canonical source:  http://web.chad.org/projects/lddaap/


Licenced under the GNU GPL v2
-----------------------------

The GNU General Public License is a Free Software license. Like any Free
Software license, it grants to you the four following freedoms:

0. The freedom to run the program for any purpose.
1. The freedom to study how the program works and adapt it to your needs.
2. The freedom to redistribute copies so you can help your neighbor.
3. The freedom to improve the program and release your improvements to the
   public, so that the whole community benefits.

You may exercise the freedoms specified here provided that you comply with the
express conditions of this license. The principal conditions are:

*  You must conspicuously and appropriately publish on each copy distributed an
   appropriate copyright notice and disclaimer of warranty and keep intact all
   the notices that refer to this License and to the absence of any warranty;
   and give any other recipients of the Program a copy of the GNU General
   Public License along with the Program. Any translation of the GNU General
   Public License must be accompanied by the GNU General Public License.

*  If you modify your copy or copies of the program or any portion of it, or
   develop a program based upon it, you may distribute the resulting work
   provided you do so under the GNU General Public License. Any translation of
   the GNU General Public License must be accompanied by the GNU General Public
   License.

*  If you copy or distribute the program, you must accompany it with the
   complete corresponding machine-readable source code or with a written offer,
   valid for at least three years, to furnish the complete corresponding
   machine-readable source code.

Any of the above conditions can be waived if you get permission from the
copyright holder.
"""



import sys
import socket
import threading
from time import sleep, asctime
from select import select
from errno import ECONNRESET, EAGAIN



def shut_down(name, exception):
	if name: print name
	global is_running
	is_running = False


def print_usage(error=None):
	print "Parameters: remote_host [remote_port [local_port]] "
	if error: 
		print error
		sys.exit(1)


class Zeroconf:

	def register_service(self, sdRef, flags, interfaceIndex, name, regtype, domain, host, listen_port, txtLen, txtRecord, callback, context):
		raise NotImplementedError

	def request_daemon_handle(self):
		raise NotImplementedError

	def unregister_service(self, sdRef):
		raise NotImplementedError

	def _make_zeroconf_txt_data(self, dictionary):
		"""Create a raw DNS TXT record.  Chad figured out the format from
		dumping his network, so this is not guaranteed to be right.  It's a
		segment of bytes, with a byte indicating the number of bytes after it
		to consider a unit of text.  The zeroconf standard expects key/value
		pairs, with the key being at least one byte that is not ASCII '=' and
		the k/v to be separated by an ASCII '='."""
		parts = []
		for k, v in dictionary.iteritems():
			t = "%s=%s" % (k, v,)
			l = len(t)
			assert l < 256, "I have no idea what happens here."
			parts.append(chr(l) + t)
		t = "".join(parts)
		return t, len(t)

	def listen_and_respond_to_zeroconf(self, listen_port, server_name):
		global is_running

		name = "%s Music proxied" % (server_name,)
		regtype = "_daap._tcp"

		sdRef = self.request_daemon_handle()
		flags = 0
		interfaceIndex = 0
		domain = 'local'
		host = ''
		txtRecord, txtLen = self._make_zeroconf_txt_data({
			"txtvers":"1",
			"Database ID":"".join([hex(ord(c)) for c in "%8.8s" % (server_name)]),
			"Machine Name":name,
			"Version":"196610",
			"Password":"false",
			"Machine ID":"".join([hex(ord(c)) for c in "%4.4s" % (server_name)]),
			"proxy-version":"0.1",
			"proxy-info":"http://web.chad.org/projects/lddaap/",
		})
		context = None

		if not self.register_service(sdRef, flags, interfaceIndex, name, regtype, domain, host, listen_port, txtLen, txtRecord, (lambda *x: 1), context):
			return

		print "Announcing %r %r via zeroconf." % (name, regtype,)
		while is_running:
			sleep(1)

		print "Removing %r %r from zeroconf." % (name, regtype,)
		self.unregister_service(sdRef)


class ComAppleBonjour(Zeroconf):

	def __init__(self):
		try:
			import bonjour
		except ImportError:
			print "See the README and prereq.sh"
			raise
		self.bonjour = bonjour

	def request_daemon_handle(self):
		return self.bonjour.AllocateDNSServiceRef()

	def register_service(self, sdRef, flags, interfaceIndex, name, regtype, domain, host, listen_port, txtLen, txtRecord, callback, context):
		ret = self.bonjour.pyDNSServiceRegister(sdRef, flags, interfaceIndex, name, regtype, domain, host, socket.htons(listen_port), txtLen, txtRecord, callback, context)

		# Error check for immediate failure
		if ret != self.bonjour.kDNSServiceErr_NoError:
			shut_down("bonjour error %r" % (ret,), None)
			return None

		return True

	def unregister_service(self, sdRef):
		self.bonjour.DNSServiceRefDeallocate(sdRef)


class OrgFreedesktopAvahi(Zeroconf):
	"""Not yet implemented."""
	pass


def listen_and_work_for_clients(listen_host, listen_port, server_name, server_port):
	global is_running

	listening_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
	listening_socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
	listening_socket.bind(("", listen_port))
	listening_socket.listen(1)
	listening_socket.settimeout(1)
	sessions = []

	print "Awaiting connections."
	while is_running:
		try:
			client_connection, client_address = listening_socket.accept()
		except socket.timeout:
			continue

		session = threading.Thread(target=handle_single_client, args=(client_connection, client_address, server_name, server_port), name="proxy %r<-proxy->%r" % ((listen_host, listen_port,),(server_name, server_port,),))
		session.start()
		sessions.append(session)


	print "Quitting proxy factory."
	for session in sessions:
		if session.isAlive(): session.join()


def handle_single_client(client_connection, client_address, server_name, server_port):
	client_connection = client_connection
	server_connection = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
	server_connection.connect((server_name, server_port))
	print "Connecting %r <-> me <-> %r." % (client_address, (server_name, server_port),)

	error = []
	while not error and is_running:

		try:
			input, void, void = select([server_connection, client_connection], [], [server_connection, client_connection], 5)
			void, output, void = select([], [server_connection, client_connection], [], 5)

			if not input: continue

			# FIXME.  This needs optimizing.  Maybe use C?
			for input_item in input:
				for output_item in output:
					if input_item == output_item: continue
					n = output_item.send(input_item.recv(4096))

					if n == 0: error = True

		except socket.error, e:
			errno = e.args[0]
			if errno == ECONNRESET:
				error = True
			elif errno == EAGAIN:
				pass

	print "Disconnecting %r from %r and local." % (client_address, (server_name, server_port),)
	server_connection.close()
	client_connection.close()


# set defaults
daap_port = socket.getservbyname("daap", "tcp")
proxied_port = daap_port
listen_port = daap_port + 10000

# get runtime parameters
try:
	proxied_host = sys.argv[1]
except IndexError:
	print_usage(error="Hostname required")

try:
	proxied_port = socket.getservbyname(sys.argv[2], "tcp")
	listen_port = socket.getservbyname(sys.argv[3], "tcp")
except IndexError:
	pass

is_running = True
proxy_thread = threading.Thread(target=listen_and_work_for_clients, args=("", listen_port, proxied_host, proxied_port), name="proxy factory")
proxy_thread.start()

zeroconf_implementation = ComAppleBonjour()
zeroconf_thread = threading.Thread(target=zeroconf_implementation.listen_and_respond_to_zeroconf, args=(listen_port, proxied_host), name="zeroconf advertizing")
zeroconf_thread.start()

try:
	while is_running:
		sleep(1)

except KeyboardInterrupt:
	is_running = False
	print "Shutting down..."
	while proxy_thread.isAlive() or zeroconf_thread.isAlive():
		proxy_thread.join(1)
		zeroconf_thread.join(1)

print "Bye."
